import 'package:example/todo/model/model.dart';
import 'package:meta/meta.dart';
import 'package:mvu/mvu.dart';
import 'package:uuid/uuid.dart';

import 'actions.dart';
import 'state.dart';

final uuid = Uuid();

const todoUpdate = TodoUpdate._();

class TodoUpdate {
  const TodoUpdate._();
  TodoState call(TodoState previousState, dynamic action) {
    print("’[ACTION] $action");
    if (action is TodoListAction) {
      final newState = action.map(
        loading: (action) => _loading(previousState, action),
        loadingSucceeded: (action) => _loadingSucceeded(previousState, action),
        loadingFailed: (action) => _loadingFailed(previousState, action),
        saving: (action) => _saving(previousState, action),
        savingFailed: (action) => _savingFailed(previousState, action),
        savingSucceeded: (action) => _savingSucceeded(previousState, action),
        updateItem: (action) => _updateItem(previousState, action),
        deleteItem: (action) => _deleteItem(previousState, action),
        createItem: (action) => _createItem(previousState, action),
      );

      print("’[NEW STATE]\n $newState");

      return newState;
    }

    return previousState;
  }

  TodoState _createItem(TodoState previousState, CreateItemAction action) {
    List<TodoItem> create(List<TodoItem> items) {
      return [
        TodoItem(
          id: uuid.v1(),
          title: action.title,
          isChecked: false,
        ),
        ...items,
      ];
    }

    return previousState.copyWith(
      list: previousState.list.maybeMap(
        loaded: (state) => state.copyWith(items: create(state.items)),
        saved: (state) => state.copyWith(items: create(state.items)),
        orElse: () => previousState.list,
      ),
    );
  }

  TodoState _updateItem(TodoState previousState, UpdateItemAction action) {
    List<TodoItem> update(List<TodoItem> items) {
      return [
        ...items.map((item) => action.itemId != item.id
            ? item
            : item.copyWith(isChecked: action.isChecked))
      ];
    }

    return previousState.copyWith(
      list: previousState.list.maybeMap(
        loaded: (state) => state.copyWith(items: update(state.items)),
        saved: (state) => state.copyWith(items: update(state.items)),
        orElse: () => previousState.list,
      ),
    );
  }

  TodoState _deleteItem(TodoState previousState, DeleteItemAction action) {
    List<TodoItem> delete(List<TodoItem> items) {
      return [
        ...items.where((item) => action.itemId != item.id),
      ];
    }

    return previousState.copyWith(
      list: previousState.list.maybeMap(
        loaded: (state) => state.copyWith(items: delete(state.items)),
        saved: (state) => state.copyWith(items: delete(state.items)),
        orElse: () => previousState.list,
      ),
    );
  }

  TodoState _loading(TodoState previousState, LoadingAction action) {
    return previousState.copyWith(
      list: previousState.list.maybeMap(
        notLoaded: (state) => TodoListState.loading(),
        orElse: () => previousState.list,
      ),
    );
  }

  TodoState _loadingSucceeded(
      TodoState previousState, LoadingSucceededAction action) {
    return previousState.copyWith(
      list: previousState.list.maybeMap(
        loading: (state) => TodoListState.loaded(
          items: action.items,
        ),
        orElse: () => previousState.list,
      ),
    );
  }

  TodoState _savingSucceeded(
      TodoState previousState, SavingSucceededAction action) {
    return previousState.copyWith(
      list: previousState.list.maybeMap(
        saving: (state) => TodoListState.saved(
          date: DateTime.now(),
          items: state.items,
        ),
        orElse: () => previousState.list,
      ),
    );
  }

  TodoState _loadingFailed(
      TodoState previousState, LoadingFailedAction action) {
    return previousState.copyWith(
      list: previousState.list.maybeMap(
        loading: (state) => TodoListState.loadingFailed(
          error: action.error,
        ),
        orElse: () => previousState.list,
      ),
    );
  }

  TodoState _saving(TodoState previousState, SavingAction action) {
    return previousState.copyWith(
      list: previousState.list.maybeMap(
        saved: (state) => TodoListState.saving(
          items: state.items,
        ),
        loaded: (state) => TodoListState.saving(
          items: state.items,
        ),
        orElse: () => previousState.list,
      ),
    );
  }

  TodoState _savingFailed(TodoState previousState, SavingFailedAction action) {
    return previousState.copyWith(
      list: previousState.list.maybeMap(
        saving: (state) => TodoListState.savingFailed(
          error: action.error,
        ),
        orElse: () => previousState.list,
      ),
    );
  }

  static ExecuteAsynchronousAction<TodoState> createItemAndSave({
    @required String title,
  }) {
    return (model, dispatch) async {
      dispatch(
        TodoListAction.createItem(title: title),
      );
      dispatch(TodoListAction.saving());
      try {
        final storage = model().configuration.storage;
        await model().list.maybeMap(
              saving: (state) => storage.save(state.items),
              orElse: () => Future.value(),
            );
        await Future.delayed(const Duration(seconds: 1));
        dispatch(
          TodoListAction.savingSucceeded(),
        );
      } catch (e) {
        dispatch(
          TodoListAction.savingFailed(error: e),
        );
      }
    };
  }

  static ExecuteAsynchronousAction<TodoState> updateItemAndSave({
    @required String itemId,
    @required bool isChecked,
  }) =>
      (model, dispatch) async {
        dispatch(
          TodoListAction.updateItem(
            itemId: itemId,
            isChecked: isChecked,
          ),
        );
        dispatch(TodoListAction.saving());
        try {
          final storage = model().configuration.storage;
          await model().list.maybeMap(
                saving: (state) => storage.save(state.items),
                orElse: () => Future.value(),
              );
          await Future.delayed(const Duration(seconds: 1));
          dispatch(
            TodoListAction.savingSucceeded(),
          );
        } catch (e) {
          dispatch(
            TodoListAction.savingFailed(error: e),
          );
        }
      };

  static ExecuteAsynchronousAction<TodoState> load() =>
      (model, dispatch) async {
        dispatch(TodoListAction.loading());
        try {
          await Future.delayed(const Duration(seconds: 1));
          final items = await model().configuration.storage.load();
          dispatch(
            TodoListAction.loadingSucceeded(items: items),
          );
        } catch (e) {
          dispatch(
            TodoListAction.loadingFailed(error: e),
          );
        }
      };
}
