import 'package:example/todo/model/model.dart';
import 'package:example/todo/state/actions.dart';
import 'package:example/todo/state/state.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:mvu/mvu.dart';

import 'create_item.dart';
import 'delete_dialog.dart';

class TodoListPage extends StatefulWidget {
  const TodoListPage({
    Key key,
  }) : super(key: key);

  @override
  _TodoListPageState createState() => _TodoListPageState();
}

class _TodoListPageState extends State<TodoListPage> {
  @override
  void initState() {
    WidgetsBinding.instance.addPostFrameCallback((timeStamp) {
      context.dispatch(TodoListAction.load());
    });
    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        title: TodoListAppBarTitle(),
      ),
      floatingActionButton: CreateFloatingButton(),
      body: View(
        selector: (TodoState state) => state.list,
        builder: (context, listState) => listState.map(
          notLoaded: (state) => LoadingView(),
          loading: (state) => LoadingView(),
          loaded: (state) => TodoListView(
            isSaving: false,
            items: state.items,
          ),
          loadingFailed: (state) => ErrorView(
            error: state.error,
          ),
          saving: (state) => TodoListView(
            isSaving: true,
            items: state.items,
          ),
          saved: (state) => TodoListView(
            isSaving: false,
            items: state.items,
          ),
          savingFailed: (state) => ErrorView(
            error: state.error,
          ),
        ),
      ),
    );
  }
}

class TodoListAppBarTitle extends StatelessWidget {
  const TodoListAppBarTitle({
    Key key,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return View(
      selector: (TodoState state) => state.list.maybeMap(
        saving: (state) => 'Saving ...',
        orElse: () => 'MVU',
      ),
      builder: (context, title) => Text(title),
    );
  }
}

class ErrorView extends StatelessWidget {
  final dynamic error;
  const ErrorView({
    Key key,
    @required this.error,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return Center(child: Text(error.toString()));
  }
}

class LoadingView extends StatelessWidget {
  const LoadingView({
    Key key,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return Center(
      child: CircularProgressIndicator(),
    );
  }
}

class TodoListView extends StatelessWidget {
  final bool isSaving;
  final List<TodoItem> items;
  const TodoListView({
    Key key,
    @required this.isSaving,
    @required this.items,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return ListView(
      children: [
        ...items.map(
          (item) => TodoItemView(
            item: item,
            isEnabled: !isSaving,
          ),
        )
      ],
    );
  }
}

class TodoItemView extends StatelessWidget {
  final TodoItem item;
  final bool isEnabled;
  TodoItemView({
    @required this.item,
    @required this.isEnabled,
  }) : super(key: Key(item.id));

  @override
  Widget build(BuildContext context) {
    return Card(
      key: ValueKey(item.id),
      child: ListTile(
        title: Text(item.title),
        onLongPress: () async {
          final result = await DeleteItemDialog.show(context);
          if (result) {
            context.dispatch(
              TodoListAction.deleteItem(
                itemId: item.id,
              ),
            );
          }
        },
        trailing: Switch(
          value: item.isChecked,
          onChanged: !isEnabled
              ? null
              : (isChecked) => context.dispatch(
                    TodoListAction.updateItemAndSave(
                      itemId: item.id,
                      isChecked: isChecked,
                    ),
                  ),
        ),
      ),
    );
  }
}

class CreateFloatingButton extends StatelessWidget {
  const CreateFloatingButton({
    Key key,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return View(
      selector: (TodoState state) {
        void create() async {
          final result = await CreateTodoPage.show(context);
          if (result != null) {
            context.dispatch(
              TodoListAction.createItemAndSave(title: result),
            );
          }
        }

        return state.list.maybeMap(
            loaded: (state) => create,
            saved: (state) => create,
            orElse: () => null);
      },
      builder: (context, onPressed) => FloatingActionButton(
        onPressed: onPressed,
        tooltip: 'Create',
        child: Icon(Icons.add),
      ),
    );
  }
}
