import 'package:flutter/widgets.dart';
import 'package:mvu/mvu.dart';

import 'action.dart';

typedef Model Update<Model, Action>(
  Model previousModel,
  Action action,
);

abstract class StoreReducer<Model, ReducerAction> {
  Type get actionType => ReducerAction;

  const StoreReducer();

  bool isActionSupported(dynamic action) => action is ReducerAction;

  bool reduce(
    Getter<Model> model,
    ReducerAction action,
    Dispatcher dispatch,
  );
}

typedef bool Reduce<Model, ReducerAction>(
  Getter<Model> model,
  ReducerAction action,
  Dispatcher dispatch,
);

class DelegateStoreReducer<Model, ReducerAction>
    extends StoreReducer<Model, ReducerAction> {
  final Reduce<Model, ReducerAction> _reduce;

  DelegateStoreReducer(this._reduce);

  @override
  bool reduce(model, ReducerAction action, dispatch) =>
      _reduce(model, action, dispatch);
}

class MvuStore<Model, Action> extends ChangeNotifier {
  final Update<Model, Action> update;
  final List<StoreReducer> _reducers = [];

  MvuStore({
    @required Model initialState,
    @required this.update,
    List<StoreReducer> reducers = const <StoreReducer>[],
  })  : assert(reducers != null),
        assert(update != null),
        this._state = initialState {
    this._reducers.addAll(
      [
        ...reducers,
        DelegateStoreReducer<Model, AsynchronousAction<Model>>(_async),
        DelegateStoreReducer<Model, Action>(_reduce),
      ],
    );
  }

  Model _state;
  Model get state => _state;

  bool _reduce(
    Getter<Model> model,
    Action action,
    Dispatcher dispatch,
  ) {
    _state = update(_state, action);
    notifyListeners();
    return true;
  }

  bool _async(
    Getter<Model> model,
    AsynchronousAction<Model> action,
    Dispatcher dispatch,
  ) {
    action.execute(() => _state, dispatch);
    return true;
  }

  Future<void> dispatch(dynamic action) async {
    final reducers = _reducers.where((x) => x.isActionSupported(action));
    for (var reducer in reducers) {
      if (reducer.reduce(() => _state, action, dispatch)) {
        break;
      }
    }
  }
}
